from itertools import chain
import numpy as np
from sklearn.base import OutlierMixin, _fit_context
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.utils.metaestimators import available_if
from sklearn.neural_network._multilayer_perceptron import (
    BaseMultilayerPerceptron,
    _STOCHASTIC_SOLVERS,
)
from sklearn.neural_network._base import DERIVATIVES


def hypersphere_loss(outputs, scores, center=1, nu=0.01, radius=0):
    loss = radius**2 + (1 / nu) * np.mean(
        np.hstack((np.zeros(scores.shape), scores)).max(1)
    )
    return loss


class OneClassMLP(OutlierMixin, BaseMultilayerPerceptron):
    """
    Unsupervised One-Class neural network. Can be used to project n-dimensional data into
    a single dimension and for outlier detection. Also used to calculate the $\\alpha$
    and $\\beta$ supports for the performance metrics used to assess synthetic data
    quality.

    Uses the loss function proposed in [1]_, which is optimized using LBFGS, stochastic
    gradient descent, or adam. This implementation refers to the architecture described
    in [2]_ using the soft-boundary objective.

    Parameters
    ----------
    hidden_layer_sizes : array-like of shape(n_layers - 2,), default=(500, 500)
        The ith element represents the number of neurons in the ith
        hidden layer.

    activation : {'identity', 'logistic', 'tanh', 'relu'}, default='relu'
        Activation function for the hidden layer.

        - 'identity', no-op activation, useful to implement linear bottleneck,
          returns f(x) = x

        - 'logistic', the logistic sigmoid function,
          returns f(x) = 1 / (1 + exp(-x)).

        - 'tanh', the hyperbolic tan function,
          returns f(x) = tanh(x).

        - 'relu', the rectified linear unit function,
          returns f(x) = max(0, x)

    nu : float, default=0.01
        Hyperparameter used in the loss function.

    solver : {'lbfgs', 'sgd', 'adam'}, default='adam'
        The solver for weight optimization.

        - 'lbfgs' is an optimizer in the family of quasi-Newton methods.

        - 'sgd' refers to stochastic gradient descent.

        - 'adam' refers to a stochastic gradient-based optimizer proposed
          by Kingma, Diederik, and Jimmy Ba

        Note: The default solver 'adam' works pretty well on relatively
        large datasets (with thousands of training samples or more) in terms of
        both training time and validation score.
        For small datasets, however, 'lbfgs' can converge faster and perform
        better.

    alpha : float, default=0.0001
        Strength of the L2 regularization term. The L2 regularization term
        is divided by the sample size when added to the loss.

    batch_size : int, default=64
        Size of minibatches for stochastic optimizers.
        If the solver is 'lbfgs', the classifier will not use minibatch.
        When set to "auto", `batch_size=min(200, n_samples)`.

    learning_rate : {'constant', 'invscaling', 'adaptive'}, default='constant'
        Learning rate schedule for weight updates.

        - 'constant' is a constant learning rate given by
          'learning_rate_init'.

        - 'invscaling' gradually decreases the learning rate at each
          time step 't' using an inverse scaling exponent of 'power_t'.
          effective_learning_rate = learning_rate_init / pow(t, power_t)

        - 'adaptive' keeps the learning rate constant to
          'learning_rate_init' as long as training loss keeps decreasing.
          Each time two consecutive epochs fail to decrease training loss by at
          least tol, the current learning rate is divided by 5.

        Only used when ``solver='sgd'``.

    learning_rate_init : float, default=1e-5
        The initial learning rate used. It controls the step-size
        in updating the weights. Only used when solver='sgd' or 'adam'.

    power_t : float, default=0.5
        The exponent for inverse scaling learning rate.
        It is used in updating effective learning rate when the learning_rate
        is set to 'invscaling'. Only used when solver='sgd'.

    max_iter : int, default=200
        Maximum number of iterations. The solver iterates until convergence
        (determined by 'tol') or this number of iterations. For stochastic
        solvers ('sgd', 'adam'), note that this determines the number of epochs
        (how many times each data point will be used), not the number of
        gradient steps.

    shuffle : bool, default=True
        Whether to shuffle samples in each iteration. Only used when
        solver='sgd' or 'adam'.

    random_state : int, RandomState instance, default=None
        Determines random number generation for weights and bias
        initialization, train-test split if early stopping is used, and batch
        sampling when solver='sgd' or 'adam'.
        Pass an int for reproducible results across multiple function calls.

    tol : float, default=0
        Tolerance for the optimization. When the loss or score is not improving
        by at least ``tol`` for ``n_iter_no_change`` consecutive iterations,
        unless ``learning_rate`` is set to 'adaptive', convergence is
        considered to be reached and training stops.

    verbose : bool, default=False
        Whether to print progress messages to stdout.

    warm_start : bool, default=False
        When set to True, reuse the solution of the previous
        call to fit as initialization, otherwise, just erase the
        previous solution.

    momentum : float, default=0.9
        Momentum for gradient descent update. Should be between 0 and 1. Only
        used when solver='sgd'.

    nesterovs_momentum : bool, default=True
        Whether to use Nesterov's momentum. Only used when solver='sgd' and
        momentum > 0.

    beta_1 : float, default=0.9
        Exponential decay rate for estimates of first moment vector in adam,
        should be in [0, 1). Only used when solver='adam'.

    beta_2 : float, default=0.999
        Exponential decay rate for estimates of second moment vector in adam,
        should be in [0, 1). Only used when solver='adam'.

    epsilon : float, default=1e-8
        Value for numerical stability in adam. Only used when solver='adam'.

    n_iter_no_change : int, default=15
        Maximum number of epochs to not meet ``tol`` improvement.
        Only effective when solver='sgd' or 'adam'.

    max_fun : int, default=15000
        Only used when solver='lbfgs'. Maximum number of loss function calls.
        The solver iterates until convergence (determined by 'tol'), number
        of iterations reaches max_iter, or this number of loss function calls.
        Note that number of loss function calls will be greater than or equal
        to the number of iterations for the `OneClassMLP`.

    Attributes
    ----------

    center_ : float
        The center of the euclidean ball used to calculate the loss.

    radius_ : float
        The radius of the euclidean ball used to calculate the loss.

    loss_ : float
        The current loss computed with the loss function.

    best_loss_ : float or None
        The minimum loss reached by the solver throughout fitting.

    loss_curve_ : list of shape (`n_iter_`,)
        The ith element in the list represents the loss at the ith iteration.

    t_ : int
        The number of training samples seen by the solver during fitting.

    coefs_ : list of shape (n_layers - 1,)
        The ith element in the list represents the weight matrix corresponding
        to layer i.

    intercepts_ : list of shape (n_layers - 1,)
        The ith element in the list represents the bias vector corresponding to
        layer i + 1.

    n_features_in_ : int
        Number of features seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during `fit`. Defined only when `X`
        has feature names that are all strings.

    n_iter_ : int
        The number of iterations the solver has run.

    n_layers_ : int
        Number of layers.

    n_outputs_ : int
        Number of outputs.

    out_activation_ : str
        Name of the output activation function.

    References
    ----------

    .. [1] Scholkopf, B., Platt, J. C., Shawe-Taylor, J., Smola, A. J., and Williamson,
        R. C. Estimating the support of a high- dimensional distribution. Neural
        computation, 13(7): 1443–1471, 2001.

    .. [2] Ruff, L., Vandermeulen, R., Goernitz, N., Deecke, L., Siddiqui, S. A., Binder,
        A., Muller, E., and Kloft, M. Deep one-class classification. In International
        conference on machine learning, pp. 4393–4402. PMLR, 2018.
    """

    def __init__(
        self,
        hidden_layer_sizes=(500, 500),
        activation="relu",
        nu=0.01,
        *,
        solver="adam",
        alpha=0.0001,
        batch_size=64,
        learning_rate="constant",
        learning_rate_init=1e-5,
        power_t=0.5,
        max_iter=200,
        shuffle=True,
        random_state=None,
        tol=0,
        verbose=False,
        warm_start=False,
        momentum=0.9,
        nesterovs_momentum=True,
        beta_1=0.9,
        beta_2=0.999,
        epsilon=1e-8,
        n_iter_no_change=15,
        max_fun=15000,
    ):
        super().__init__(
            hidden_layer_sizes=hidden_layer_sizes,
            activation=activation,
            solver=solver,
            alpha=alpha,
            batch_size=batch_size,
            learning_rate=learning_rate,
            learning_rate_init=learning_rate_init,
            power_t=power_t,
            max_iter=max_iter,
            loss=None,
            shuffle=shuffle,
            random_state=random_state,
            tol=tol,
            verbose=verbose,
            warm_start=warm_start,
            momentum=momentum,
            nesterovs_momentum=nesterovs_momentum,
            early_stopping=False,
            validation_fraction=None,
            beta_1=beta_1,
            beta_2=beta_2,
            epsilon=epsilon,
            n_iter_no_change=n_iter_no_change,
            max_fun=max_fun,
        )
        self.nu = nu

    def _init_center(self, y_pred):
        """
        Initialize hypersphere center c as the mean from an initial forward pass on the
        data.
        """
        self.center_ = y_pred.mean()

    def _init_radius(self, y_pred):
        dist = np.sum((y_pred - self.center_) ** 2, axis=1).reshape(-1, 1)
        self.radius_ = np.quantile(np.sqrt(dist), 1 - self.nu)

    def _initialize(self, layer_units, dtype):
        # set all attributes, allocate weights etc. for first call
        # Initialize parameters
        self.n_iter_ = 0
        self.t_ = 0

        # Compute the number of layers
        self.n_layers_ = len(layer_units)

        # Output for regression
        self.out_activation_ = "identity"

        # Initialize coefficient and intercept layers
        self.coefs_ = []
        self.intercepts_ = []

        for i in range(self.n_layers_ - 1):
            coef_init, intercept_init = self._init_coef(
                layer_units[i], layer_units[i + 1], dtype
            )
            self.coefs_.append(coef_init)
            self.intercepts_.append(intercept_init)

        if self.solver in _STOCHASTIC_SOLVERS:
            self.loss_curve_ = []
            self._no_improvement_count = 0
            self.best_loss_ = np.inf

    def _backprop(self, X, y, activations, deltas, coef_grads, intercept_grads):
        """Compute the MLP loss function and its corresponding derivatives
        with respect to each parameter: weights and bias vectors.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        y : ndarray of shape (n_samples,)
            The target values.

        activations : list, length = n_layers - 1
             The ith element of the list holds the values of the ith layer.

        deltas : list, length = n_layers - 1
            The ith element of the list holds the difference between the
            activations of the i + 1 layer and the backpropagated error.
            More specifically, deltas are gradients of loss with respect to z
            in each layer, where z = wx + b is the value of a particular layer
            before passing through the activation function

        coef_grads : list, length = n_layers - 1
            The ith element contains the amount of change used to update the
            coefficient parameters of the ith layer in an iteration.

        intercept_grads : list, length = n_layers - 1
            The ith element contains the amount of change used to update the
            intercept parameters of the ith layer in an iteration.

        Returns
        -------
        loss : float
        coef_grads : list, length = n_layers - 1
        intercept_grads : list, length = n_layers - 1
        """
        n_samples = X.shape[0]

        # Forward propagate
        activations = self._forward_pass(activations)

        dist = np.sum((activations[-1] - self.center_) ** 2, axis=1).reshape(-1, 1)
        scores = dist - self.radius_**2

        # Get loss
        loss = hypersphere_loss(
            outputs=activations[-1],
            scores=scores,
            center=self.center_,
            nu=self.nu,
            radius=self.radius_,
        )

        # Backward propagate
        last = self.n_layers_ - 2

        # The calculation of delta[last] here works with following
        # combinations of output activation and loss function:
        # sigmoid and binary cross entropy, softmax and categorical cross
        # entropy, and identity with squared loss
        deltas[last] = scores

        # Compute gradient for the last layer
        self._compute_loss_grad(
            last, n_samples, activations, deltas, coef_grads, intercept_grads
        )

        inplace_derivative = DERIVATIVES[self.activation]
        # Iterate over the hidden layers
        for i in range(self.n_layers_ - 2, 0, -1):
            deltas[i - 1] = safe_sparse_dot(deltas[i], self.coefs_[i].T)
            inplace_derivative(activations[i], deltas[i - 1])

            self._compute_loss_grad(
                i - 1, n_samples, activations, deltas, coef_grads, intercept_grads
            )

        return loss, coef_grads, intercept_grads

    @available_if(lambda est: est._check_solver())
    @_fit_context(prefer_skip_nested_validation=True)
    def partial_fit(self, X, y=None):
        """Update the model with a single iteration over the given data.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        y : ndarray of shape (n_samples,)
            The target values.

        Returns
        -------
        self : object
            Trained MLP model.
        """
        return self._fit(X, incremental=True)

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y=None):
        """Fit the model to data matrix X.

        Parameters
        ----------
        X : ndarray or sparse matrix of shape (n_samples, n_features)
            The input data.

        Returns
        -------
        self : object
            Returns a trained MLP model.
        """
        return self._fit(X, incremental=False)

    def _fit(self, X, incremental=False):
        # Make sure self.hidden_layer_sizes is a list
        hidden_layer_sizes = self.hidden_layer_sizes
        if not hasattr(hidden_layer_sizes, "__iter__"):
            hidden_layer_sizes = [hidden_layer_sizes]
        hidden_layer_sizes = list(hidden_layer_sizes)

        if np.any(np.array(hidden_layer_sizes) <= 0):
            raise ValueError(
                "hidden_layer_sizes must be > 0, got %s." % hidden_layer_sizes
            )
        first_pass = not hasattr(self, "coefs_") or (
            not self.warm_start and not incremental
        )

        X = self._validate_input(X, incremental, reset=first_pass)

        n_samples, n_features = X.shape

        self.n_outputs_ = 1

        layer_units = [n_features] + hidden_layer_sizes + [self.n_outputs_]

        # check random state
        self._random_state = check_random_state(self.random_state)

        if first_pass:
            # First time training the model
            self._initialize(layer_units, X.dtype)

            y_pred = self._forward_pass_fast(X, check_input=False)
            self._init_center(y_pred)
            self._init_radius(y_pred)

        # Initialize lists
        activations = [X] + [None] * (len(layer_units) - 1)
        deltas = [None] * (len(activations) - 1)

        coef_grads = [
            np.empty((n_fan_in_, n_fan_out_), dtype=X.dtype)
            for n_fan_in_, n_fan_out_ in zip(layer_units[:-1], layer_units[1:])
        ]

        intercept_grads = [
            np.empty(n_fan_out_, dtype=X.dtype) for n_fan_out_ in layer_units[1:]
        ]

        # Run the Stochastic optimization solver
        if self.solver in _STOCHASTIC_SOLVERS:
            self._fit_stochastic(
                X,
                np.ones(X.shape[0]),
                activations,
                deltas,
                coef_grads,
                intercept_grads,
                layer_units,
                incremental,
            )

        # Run the LBFGS solver
        elif self.solver == "lbfgs":
            self._fit_lbfgs(
                X,
                np.ones(X.shape[0]),
                activations,
                deltas,
                coef_grads,
                intercept_grads,
                layer_units,
            )

        # validate parameter weights
        weights = chain(self.coefs_, self.intercepts_)
        if not all(np.isfinite(w).all() for w in weights):
            raise ValueError(
                "Solver produced non-finite parameter weights. The input data may"
                " contain large values and need to be preprocessed."
            )

        return self

    def _validate_input(self, X, incremental, reset):
        X = self._validate_data(
            X,
            accept_sparse=["csr", "csc"],
            dtype=(np.float64, np.float32),
            reset=reset,
        )
        return X

    def predict(self, X):
        """Predict using the multi-layer perceptron model.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input data.

        Returns
        -------
        y : ndarray of shape (n_samples, n_outputs)
            The predicted values.
        """
        check_is_fitted(self)
        return self._predict(X)

    def _predict(self, X, check_input=True):
        """Private predict method with optional input validation"""
        scores = self._forward_pass_fast(X, check_input=check_input)
        return scores.ravel()



from itertools import product
from sklearn.base import clone
from sklearn.utils import check_random_state
from sklearn.preprocessing import FunctionTransformer
from sklearn.model_selection import ParameterGrid
from imblearn.utils import Substitution
from imblearn.pipeline import Pipeline
from imblearn.utils._docstring import _random_state_docstring


@Substitution(
    random_state=_random_state_docstring,
)
def check_random_states(random_state, n_runs):
    """
    Create random states for experiments. Used to create seeds for different
    initializations.

    Parameters
    ----------
    {random_state}

    n_runs : int
        Number of initializations.

    Returns
    -------
    random_states : list
        A list of random states with length ``n_runs``.
    """
    random_state = check_random_state(random_state)
    return [random_state.randint(0, 2**32 - 1, dtype="uint32") for _ in range(n_runs)]


@Substitution(
    random_state=_random_state_docstring,
)
def check_pipelines(*objects_list, random_state, n_runs):
    """
    Extract estimators and parameter grids to be passed to ModelSearchCV. This enables
    searching over any sequence of parameter settings and objects.

    Parameters
    ----------
    *objects_list : sequence of lists
        Lists of objects to be chained in a pipeline in the passed order. Each list
        must contain tuples composed of (``<obj_name>``, ``<object>``,
        ``<parameter_values_dict>``).

    {random_state}

    n_runs : int
        Number of initializations.

    Returns
    -------
    estimators : List of Pipelines with all combinations among the passed lists of
        objects.

    param_grids : List of dictionaries with estimator and parameter names (``str``) as
        keys and lists of parameter settings to try as values.
    """

    # Create random states
    random_states = check_random_states(random_state, n_runs)

    pipelines = []
    param_grid = []
    for comb, rs in product(product(*objects_list), random_states):
        name = "|".join([i[0] for i in comb])

        # name, object, sub grid
        comb = [
            (
                (nm, ob, ParameterGrid(sg))
                if ob is not None
                else (nm, FunctionTransformer(), ParameterGrid(sg))
            )
            for nm, ob, sg in comb
        ]

        # Create estimator
        if name not in [n[0] for n in pipelines]:
            est = Pipeline([(nm, ob) for nm, ob, _ in comb])
            pipelines.append((name, est))

        # Create intermediate parameter grids
        sub_grids = [
            [{f"{nm}__{k}": v for k, v in param_def.items()} for param_def in sg]
            for nm, obj, sg in comb
        ]

        # Create parameter grids
        for sub_grid in product(*sub_grids):
            param_prefix = f"{name}__"
            grid = {"est_name": [name]}
            grid.update(
                {f"{param_prefix}{k}": [v] for d in sub_grid for k, v in d.items()}
            )
            random_states = {
                f"{param_prefix}{param}": [rs]
                for param in est.get_params()
                if "random_state" in param
            }
            grid.update(random_states)

            # Avoid multiple runs over pipelines without random state
            if grid not in param_grid:
                param_grid.append(grid)

    return pipelines, param_grid


def check_pipelines_wrapper(
    *objects_list,
    wrapper,
    random_state,
    n_runs,
    estimator_param="classifier",
    wrapped_only=True,
):
    """
    Extract estimators within a wrapper object and parameter grids to be passed to
    ModelSearchCV. This enables searching over any sequence of parameter settings and
    objects.

    Parameters
    ----------
    *objects_list : sequence of lists
        Lists of objects to be chained in a pipeline in the passed order. Each list
        must contain tuples composed of (``<obj_name>``, ``<object>``,
        ``<parameter_values_dict>``).

    wrapper : tuple or tuple
        Wrapper object to which the lists of objects will be passed. Must be structured
        as (``<obj_name>``, ``<object>``, ``<parameter_values_dict>``) and .

    {random_state}

    n_runs : int
        Number of initializations.

    estimator_param : str, default="classifier"
        Name of the parameter in the wrapper object where the estimators will be passed.

    wrapped_only : bool, default=True
        Return only the wrapped estimators. If ``False``, returns both the wrapped and
        the original objects.

    Returns
    -------
    wrapped_estimators : List of Pipelines with all combinations among the passed lists
        of objects.

    wrapped_param_grids : List of dictionaries with estimator and parameter names
        (``str``) as keys and lists of parameter settings to try as values.
    """

    wrapper_label = wrapper[0]
    wrapper_obj = wrapper[1]
    wrapper_grid = wrapper[2]

    estimators, param_grids = check_pipelines(
        *objects_list, random_state=random_state, n_runs=n_runs
    )

    wrapped_estimators = [
        (
            f"{wrapper_label}|{name}",
            clone(wrapper_obj).set_params(**{estimator_param: pipeline}),
        )
        for name, pipeline in estimators
    ]

    def _format_param(param):
        return "__".join(param.split("__")[1:])

    wrapped_param_grids = [
        {
            "est_name": [f'{wrapper_label}|{d["est_name"][0]}'],
            **{
                f'{wrapper_label}|{d["est_name"][0]}__{estimator_param}__'
                + f"{_format_param(k)}": v
                for k, v in d.items()
                if k != "est_name"
            },
            **{
                f'{wrapper_label}|{d["est_name"][0]}__{k}': v
                for k, v in wrapper_grid.items()
            },
        }
        for d in param_grids
    ]
    if wrapped_only:
        return wrapped_estimators, wrapped_param_grids
    else:
        return (estimators + wrapped_estimators, param_grids + wrapped_param_grids)


def check_param_grids(param_grids, est_names):
    """Check the parameters grids to use with
    parametrized estimators."""

    # Check the parameters grids
    flat_param_grids = [
        param_grid for param_grid in list(ParameterGrid(param_grids)) if param_grid
    ]

    # Append existing estimators names
    param_grids = []
    for param_grid in flat_param_grids:

        # Get estimator name
        est_name = param_grid.pop("est_name", None)

        # Modify values
        param_grid = {param: [val] for param, val in param_grid.items()}

        # Check estimators prefixes
        params_prefixes = set([param.split("__")[0] for param in param_grid.keys()])
        if not params_prefixes.issubset(est_names):
            raise ValueError(
                "Parameters prefixes are not subset of parameter `est_names`."
            )
        if len(params_prefixes) > 1:
            raise ValueError("Parameters prefixes are not unique.")
        if est_name is not None and len(params_prefixes.union([est_name])) > 1:
            raise ValueError(
                "Parameters prefixes and parameter `est_name` are not unique."
            )
        param_grid["est_name"] = (
            [est_name] if est_name is not None else list(params_prefixes)
        )

        # Append parameter grid
        param_grids.append(param_grid)

    # Append missing estimators names
    current_est_names = set([param_grid["est_name"][0] for param_grid in param_grids])
    missing_est_names = set(est_names).difference(current_est_names)
    for est_name in missing_est_names:
        param_grids.append({"est_name": [est_name]})

    return param_grids


def check_estimator_type(estimators):
    """Returns the type of estimators."""
    estimator_types = set([estimator._estimator_type for _, estimator in estimators])
    if len(estimator_types) > 1:
        raise ValueError(
            "Both classifiers and regressors were found. "
            "A single estimator type should be included."
        )
    return estimator_types.pop()



import pytest
import os
import time
import contextlib
import io
from mlresearch.utils._parallelize import _get_n_jobs, parallel_loop


def example_function(a):
    time.sleep(0.5)
    return None


def test_get_n_jobs():
    max_jobs = os.cpu_count()
    other_n_jobs = 2 if max_jobs >= 2 else 1

    assert _get_n_jobs(None) == 1
    assert _get_n_jobs(other_n_jobs) == other_n_jobs
    assert _get_n_jobs(-1) == max_jobs

    with pytest.raises(RuntimeError):
        _get_n_jobs(max_jobs + 1)


def test_parallel_loop():
    # Get number of jobs to set iterable range
    n_jobs = os.cpu_count()

    # Check if parallelization is happening
    start = time.time()
    parallel_loop(example_function, range(n_jobs), n_jobs=-1)
    exec_time = time.time() - start

    exp_single_job = n_jobs * 0.5 if n_jobs != 1 else 1
    assert exec_time < exp_single_job


def test_progress_bar():
    # Get number of jobs to set iterable range
    n_jobs = os.cpu_count()
    description = "Test descr."

    # Check if parallelization is happening
    f = io.StringIO()
    with contextlib.redirect_stderr(f):
        parallel_loop(
            function=example_function,
            iterable=range(n_jobs),
            n_jobs=-1,
            progress_bar=True,
            description=description,
        )

    output = f.getvalue().splitlines()[1:]
    assert all([out_.startswith(description) for out_ in output])



import numpy as np
from sklearn.metrics._scorer import _BaseScorer, _SCORERS
from sklearn.utils.multiclass import type_of_target
from sklearn.utils import column_or_1d


class RankingScorer(_BaseScorer):
    """
    Make an Rank-based scorer from a probability-based metric or loss function.

    This factory class wraps scoring functions to be used in
    :class:`~mlresearch.model_selection.ModelSearchCV` and
    :class:`~sklearn.model_selection.GridSearchCV`. It takes a score function, such as
    :func:`~mlresearch.metrics.precision_at_k` and is used to score a
    classifier. The signature of the call is `(estimator, X, y)` where `estimator` is the
    model to be evaluated, `X` is the data and `y` is the ground truth labeling (or
    `None` in the case of unsupervised models).

    Parameters
    ----------
    score_func : callable
        Score function (or loss function) with signature
        ``score_func(y, y_pred, **kwargs)``.

    sign : int, default=1
        Use 1 to keep the original variable's scale, use -1 to reverse the scale.

    **kwargs : dict
        Other parameters passed to the scorer. Refer to
        :func:`set_score_request` for more details.

    Returns
    -------
    scorer : callable
        Callable object that returns a scalar score.
    """

    def __init__(self, score_func, sign=1, **kwargs):
        self._score_func = score_func
        self._sign = sign
        self._kwargs = kwargs

    def _score(self, method_caller, clf, X, y, sample_weight=None):
        """Evaluate predicted probabilities for X relative to y_true.
        Parameters
        ----------
        method_caller : callable
            Returns predictions given an estimator, method name, and other
            arguments, potentially caching results.
        clf : object
            Trained classifier to use for scoring. Must have a `predict_proba`
            method; the output of that is used to compute the score.
        X : {array-like, sparse matrix}
            Test data that will be fed to clf.predict_proba.
        y : array-like
            Gold standard target values for X. These must be class labels,
            not probabilities.
        sample_weight : array-like, default=None
            Sample weights.
        Returns
        -------
        score : float
            Score function applied to prediction of estimator on X.
        """

        y_type = type_of_target(y)
        y_pred = method_caller(clf, "predict_proba", X)
        if y_type == "binary" and len(y_pred.shape) == 2 and y_pred.shape[1] <= 2:
            # `y_type` could be equal to "binary" even in a multi-class
            # problem: (when only 2 class are given to `y_true` during scoring)
            # Thus, we need to check for the shape of `y_pred`.
            target_idx = (
                self._kwargs["target_label"]
                if "target_label" in self._kwargs.keys()
                else -1
            )
            y_pred = y_pred[:, target_idx]

        if sample_weight is not None:
            return self._sign * self._score_func(
                y, y_pred, sample_weight=sample_weight, **self._kwargs
            )
        else:
            return self._sign * self._score_func(y, y_pred, **self._kwargs)

    def _factory_args(self):
        return ", needs_proba=True"

    def set_score_request(self):
        """
        Placeholder to overwrite sklearn's ``_BaseScorer.set_score_request`` function.
        It is not used and was raising a docstring error with scikit-learn v1.3.0.

        Note
        ----
        This placeholder will be removed soon
        """
        pass


def precision_at_k(y_true, y_score, k=10, target_label=1):
    """
    Calculate precision at ``k``, where ``k`` is the number of relevant items to consider
    (sorted in descending order by its score). This metric consists of the ration between
    the number of items with label ``target_label``, out of the top ``k`` items with
    highest scores.

    .. warning::
        This metric is not the same as ``sklearn.metrics.top_k_accuracy_score``, which
        calculates the amount of times ``y_true`` is within the top ``k`` predicted
        classes for each item.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        True labels.

    y_score : array-like of shape (n_samples,) or (n_samples, n_classes)
        Target scores. These can be either probability estimates or non-thresholded
        decision values (as returned by `decision_function` on some classifiers).
        Expects scores with shape (n_samples,).

    k : int, default=10
        Number of most likely predictions considered to compute the number of correct
        labels.

    target_label : int, default=1
        Value of the label with relevant items.
    """

    # Makes this compatible with various array types
    y_true_arr = column_or_1d(y_true)
    y_score_arr = column_or_1d(y_score)

    y_true_arr = y_true_arr == target_label
    top_idx = np.argsort(y_score_arr)[-k:]
    return y_true_arr[top_idx].sum() / k


_SCORERS["precision_at_10"] = RankingScorer(precision_at_k, k=10)



import os
import contextlib
from joblib import Parallel, delayed
from ._utils import _optional_import


def _get_n_jobs(n_jobs):
    """Assign number of jobs to be assigned in parallel."""
    max_jobs = os.cpu_count()
    n_jobs = 1 if n_jobs is None else int(n_jobs)
    if n_jobs > max_jobs:
        raise RuntimeError("Cannot assign more jobs than the number of CPUs.")
    elif n_jobs == -1:
        return max_jobs
    else:
        return n_jobs


@contextlib.contextmanager
def _tqdm_joblib(tqdm_object):
    """
    Context manager to patch joblib to report into tqdm progress bar given as argument.
    """

    def tqdm_print_progress(self):
        if self.n_completed_tasks > tqdm_object.n:
            n_completed = self.n_completed_tasks - tqdm_object.n
            tqdm_object.update(n=n_completed)

    original_print_progress = Parallel.print_progress
    Parallel.print_progress = tqdm_print_progress

    try:
        yield tqdm_object
    finally:
        Parallel.print_progress = original_print_progress
        tqdm_object.close()


def parallel_loop(
    function, iterable, n_jobs=None, progress_bar=False, description=None
):
    """
    Parallelize a loop and optionally add a progress bar.

    .. warning::
        The progress bar tracks job starts, not completions.

    Parameters
    ----------
    function : function
        The function to which the elements in the iterable will passed to. Must have a
        single parameter.

    iterable : iterable
        Object to be looped over.

    n_jobs : int, default=None
        Number of jobs to run in parallel. None means 1 unless in a
        joblib.parallel_backend context. -1 means using all processors.

    Returns
    -------
    output : list
        The list with the results produced using ``function`` across ``iterable``.
    """
    n_jobs = _get_n_jobs(n_jobs)

    if progress_bar:
        tqdm = _optional_import("tqdm.auto").tqdm

        with _tqdm_joblib(tqdm(desc=description, total=len(iterable))) as progress_bar:
            return Parallel(n_jobs=n_jobs)(delayed(function)(i) for i in iterable)

    else:
        return Parallel(n_jobs=n_jobs)(delayed(function)(i) for i in iterable)



"""
Test the search module.
"""

import pytest
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.datasets import make_regression, make_classification

from ...model_selection._search import (
    MultiEstimatorMixin,
    MultiRegressor,
    MultiClassifier,
    ModelSearchCV,
)


RND_SEED = 0
X_reg, y_reg = make_regression(random_state=RND_SEED)
X_clf, y_clf = make_classification(random_state=RND_SEED)
REGRESSORS = [
    ("lr", LinearRegression()),
    ("dtr", DecisionTreeRegressor()),
    ("pip", make_pipeline(StandardScaler(with_mean=False), LinearRegression())),
]
CLASSIFIERS = [
    ("lr", LogisticRegression()),
    ("dtc", DecisionTreeClassifier()),
    (
        "pip",
        make_pipeline(StandardScaler(with_mean=False), LogisticRegression()),
    ),
]
REGRESSORS_PARAM_GRIDS = [
    {"dtr__max_depth": [3, 5], "dtr__random_state": [RND_SEED, RND_SEED + 3]},
    {
        "pip__standardscaler__with_std": [True, False],
        "pip__linearregression__fit_intercept": [True, False],
    },
]
CLASSIFIERS_PARAM_GRIDS = {
    "dtc__max_depth": [3, 5],
    "dtc__criterion": ["entropy", "gini"],
    "dtc__random_state": [RND_SEED, RND_SEED + 5],
}


@pytest.mark.parametrize("estimators", [None, [], DecisionTreeClassifier()])
def test_multi_estimator_wrong_type(estimators):
    """Test the initialization of multi-estimator class with wrong inputs."""
    with pytest.raises(TypeError):
        MultiEstimatorMixin(estimators, "est").fit(X_clf, y_clf)


def test_multi_estimator_unique_names():
    """Test the initialization of multi-estimator class with duplicate names."""
    estimators = [("est", LinearRegression()), ("est", DecisionTreeRegressor())]
    with pytest.raises(ValueError):
        MultiEstimatorMixin(estimators, "estimator").fit(X_clf, y_clf)


def test_multi_estimator_wrong_name():
    """Test the initialization of multi-estimator class with wrong estimator name."""
    estimators = [("lr", LinearRegression()), ("dtr", DecisionTreeRegressor())]
    with pytest.raises(ValueError):
        MultiEstimatorMixin(estimators, "est").fit(X_clf, y_clf)


def test_multi_estimator_params_methods():
    """Test the set and get parameters methods."""

    # Get parameters
    est_name = "dtr"
    multi_estimator = MultiEstimatorMixin(REGRESSORS, est_name)
    params = multi_estimator.get_params(deep=False)
    assert params["est_name"] == "dtr"

    # Set parameters
    est_name = "reg"
    multi_estimator.set_params(est_name="reg")
    params = multi_estimator.get_params(deep=False)
    assert params["est_name"] == est_name


@pytest.mark.parametrize(
    "estimators,est_name,X,y",
    [
        (REGRESSORS, "lr", X_reg, y_reg),
        (REGRESSORS, "dtr", X_reg, y_reg),
        (CLASSIFIERS, "lr", X_clf, y_clf),
        (CLASSIFIERS, "dtc", X_clf, y_clf),
    ],
)
def test_multi_estimator_fitting(estimators, est_name, X, y):
    """Test multi-estimator fitting process."""
    multi_estimator = MultiEstimatorMixin(estimators, est_name)
    multi_estimator.fit(X, y)
    fitted_estimator = dict(estimators)[est_name]
    assert isinstance(fitted_estimator, multi_estimator.estimator_.__class__)
    assert fitted_estimator.get_params() == multi_estimator.estimator_.get_params()


@pytest.mark.parametrize(
    "estimators,X,y,est_name",
    [(REGRESSORS, X_reg, y_reg, "reg"), (CLASSIFIERS, X_clf, y_clf, None)],
)
def test_multi_estimator_fitting_error(estimators, X, y, est_name):
    """Test parametrized estimators fitting error."""
    with pytest.raises(ValueError):
        MultiEstimatorMixin(estimators, est_name).fit(X, y)


def test_multi_classifier_type():
    """Test multi-classifier type of estimator attribute."""
    multi_clf = MultiClassifier(CLASSIFIERS)
    assert multi_clf._estimator_type == "classifier"


def test_multi_regressor_type():
    """Test multi-regressor type of estimator attribute."""
    multi_reg = MultiRegressor(REGRESSORS)
    assert multi_reg._estimator_type == "regressor"


@pytest.mark.parametrize(
    "estimators,param_grids,estimator_type",
    [
        (REGRESSORS, REGRESSORS_PARAM_GRIDS, "regressor"),
        (CLASSIFIERS, CLASSIFIERS_PARAM_GRIDS, "classifier"),
    ],
)
def test_model_search_cv(estimators, param_grids, estimator_type):
    """Test model search cv."""
    est_names, *_ = zip(*estimators)
    mscv = ModelSearchCV(estimators, param_grids)
    if estimator_type == "regressor":
        mscv.fit(X_reg, y_reg)
    elif estimator_type == "classifier":
        mscv.fit(X_clf, y_clf)
    assert set(est_names) == set(mscv.cv_results_["param_est_name"])



import numpy as np
from sklearn.utils import check_random_state
from sklearn.preprocessing import MinMaxScaler
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils._testing import ignore_warnings
from mlresearch.neural_network import OneClassMLP
from mlresearch.metrics import AlphaPrecision, BetaRecall, Authenticity


rng = check_random_state(42)
X = np.concatenate([rng.normal(size=(500, 1)), rng.normal(size=(500, 1))], axis=1)
X = MinMaxScaler().fit_transform(X)
X_synth = X.copy() + 0.1
alpha = beta = 0.05


def test_alpha_precision():
    mlp_real = OneClassMLP(max_iter=5, random_state=42)
    with ignore_warnings(category=ConvergenceWarning):
        mlp_real.fit(X)

    precision = AlphaPrecision(mlp_real.predict, alpha=alpha).fit(X)
    scores = precision.score(X_synth)

    # Test __repr__ bug due to undefined attributes (#67)
    assert precision

    assert isinstance(scores, np.ndarray)
    assert scores.shape == (500,)
    assert (np.unique(scores) == np.array([0, 1])).all()


def test_beta_recall_with_scorer():
    mlp_synth = OneClassMLP(max_iter=5, random_state=42)
    with ignore_warnings(category=ConvergenceWarning):
        mlp_synth.fit(X)

    recall = BetaRecall(mlp_synth.predict, beta=beta).fit(X_synth)
    scores = recall.score(X)

    # Test __repr__ bug due to undefined attributes (#67)
    assert recall

    assert isinstance(scores, np.ndarray)
    assert scores.shape == (500,)
    assert (np.unique(scores) == np.array([0, 1])).all()


def test_beta_recall_without_scorer():
    recall = BetaRecall(beta=beta).fit(X_synth)
    scores = recall.score(X)

    # Test __repr__ bug due to undefined attributes (#67)
    assert recall

    assert isinstance(scores, np.ndarray)
    assert scores.shape == (500,)
    assert (np.unique(scores) == np.array([0, 1])).all()


def test_authenticity():
    authenticity = Authenticity().fit(X)
    scores = authenticity.score(X_synth)

    # Test __repr__ bug due to undefined attributes (#67)
    assert authenticity

    assert isinstance(scores, np.ndarray)
    assert scores.shape == (500,)
    assert (np.unique(scores) == np.array([0, 1])).all()



"""
It includes utilities to search the parameter and model space.

Extracted from the no longer maintained ``research-learn`` library.
"""

# License: BSD 3 clause

from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, clone
from sklearn.utils.metaestimators import _BaseComposition
from sklearn.utils.validation import check_is_fitted
from sklearn.model_selection import GridSearchCV

from ..utils._check_pipelines import check_estimator_type, check_param_grids


class MultiEstimatorMixin(_BaseComposition):
    """Mixin class for multi estimator."""

    def __init__(self, estimators, est_name=None):

        self.estimators = estimators
        self.est_name = est_name

    def _validate_estimators(self):
        error_msg = "Invalid `estimators` attribute, `estimators` should"
        " be a list of (string, estimator) tuples."
        try:
            if len(self.estimators) == 0:
                raise TypeError(error_msg)
            for name, est in self.estimators:
                is_str = isinstance(name, str)
                is_est = isinstance(est, BaseEstimator)
                if not (is_str and is_est):
                    raise TypeError(error_msg)
        except TypeError:
            raise TypeError(error_msg)
        self.est_names_ = [est_name for est_name, _ in self.estimators]
        super(MultiEstimatorMixin, self)._validate_names(self.est_names_)
        if self.est_name not in self.est_names_:
            raise ValueError(
                f'Attribute `est_name` should be one of {", ".join(self.est_names_)}. '
                f"Got `{self.est_name}` instead."
            )

    def set_params(self, **params):
        """Set the parameters.
        Valid parameter keys can be listed with get_params().

        Parameters
        ----------
        params : keyword arguments
            Specific parameters using e.g. set_params(parameter_name=new_value)
            In addition, to setting the parameters of the ``MultiEstimatorMixin``,
            the individual estimators of the ``MultiEstimatorMixin`` can also be
            set or replaced by setting them to None.
        """
        super(MultiEstimatorMixin, self)._set_params("estimators", **params)
        return self

    def get_params(self, deep=True):
        """Get the parameters.

        Parameters
        ----------
        deep: bool
            Setting it to True gets the various estimators and the parameters
            of the estimators as well
        """
        return super(MultiEstimatorMixin, self)._get_params("estimators", deep=deep)

    def fit(self, X, y, **fit_params):
        """Fit the selected estimator."""

        # Validate estimators
        self._validate_estimators()

        # Copy one of the estimators
        estimator = clone(dict(self.estimators)[self.est_name])

        # Fit estimator
        self.estimator_ = estimator.fit(X, y, **fit_params)

        if hasattr(estimator, "classes_"):
            self.classes_ = estimator.classes_

        return self

    def predict(self, X):
        """Predict with the selected estimator."""
        check_is_fitted(self, "estimator_")
        return self.estimator_.predict(X)


class MultiClassifier(MultiEstimatorMixin, ClassifierMixin):
    """The functionality of a collection of classifiers is provided as
    a single metaclassifier. The classifier to be fitted is selected using a
    parameter."""

    def predict_proba(self, X):
        """Predict the probability with the selected estimator."""
        check_is_fitted(self, "estimator_")
        return self.estimator_.predict_proba(X)


class MultiRegressor(MultiEstimatorMixin, RegressorMixin):
    """The functionality of a collection of regressors is provided as
    a single metaregressor. The regressor to be fitted is selected using a
    parameter."""

    pass


class ModelSearchCV(GridSearchCV):
    """Exhaustive search over specified parameter values for a collection of estimators.

    Important members are fit, predict.

    ModelSearchCV implements a "fit" and a "score" method.
    It also implements "predict", "predict_proba", "decision_function",
    "transform" and "inverse_transform" if they are implemented in the
    estimators used.

    The parameters of the estimators used to apply these methods are optimized
    by cross-validated grid-search over their parameter grids.

    Read more in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    estimators :  list of (string, estimator) tuples
        Each estimator is assumed to implement the scikit-learn estimator interface.
        Either estimator needs to provide a ``score`` function,
        or ``scoring`` must be passed.

    param_grids : dict or list of dictionaries
        Dictionary with parameters names (string) as keys and lists of
        parameter settings to try as values, or a list of such
        dictionaries, in which case the grids spanned by each dictionary
        in the list are explored. This enables searching over any sequence
        of parameter settings.

    scoring : string, callable, list/tuple, dict or None, optional (default=None)
        A single string or a callable to evaluate the predictions on the
        test set.

        For evaluating multiple metrics, either give a list of (unique) strings
        or a dict with names as keys and callables as values.

        NOTE that when using custom scorers, each scorer should return a single
        value. Metric functions returning a list/array of values can be wrapped
        into multiple scorers that return one value each.

        If ``None``, a default scorer is used.

    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.

        - When ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.

        - When ``-1`` means using all processors.

    pre_dispatch : int or string, optional (default=None)
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:

        - ``None``, in which case all the jobs are immediately created.

        - An int, giving the exact number of total jobs that are spawned.

        - A string, as a function of n_jobs i.e. ``'2*n_jobs'``.

    cv : int, cross-validation generator or an iterable, optional (defalut=None)
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - ``None``, to use the default 3-fold cross validation.

        - integer, to specify the number of folds in a ``(Stratified)KFold``.

        - An object to be used as a cross-validation generator.

        - An iterable yielding train, test splits.

    refit : boolean, string, or callable, optional (default=True)
        Refit an estimator using the best found parameters on the whole
        dataset.

        For multiple metric evaluation, this needs to be a string denoting the
        scorer that would be used to find the best parameters for refitting
        the estimator at the end.

        Where there are considerations other than maximum score in
        choosing a best estimator, ``refit`` can be set to a function which
        returns the selected ``best_index_`` given ``cv_results_``. In that
        case, the ``best_estimator_`` and ``best_parameters_`` will be set
        according to the returned ``best_index_`` while the ``best_score_``
        attribute will not be availble.

        The refitted estimator is made available at the ``best_estimator_``
        attribute and permits using ``predict`` directly on this
        ``ModelSearchCV`` instance.

        Also for multiple metric evaluation, the attributes ``best_index_``,
        ``best_score_`` and ``best_params_`` will only be available if
        ``refit`` is set and all of them will be determined w.r.t this specific
        scorer.

        See ``scoring`` parameter to know more about multiple metric
        evaluation.

    verbose : integer, optional (default=0)
        Controls the verbosity: the higher, the more messages.

    error_score : 'raise' or numeric, optional (default=np.nan)
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised. If a numeric value is given,
        FitFailedWarning is raised. This parameter does not affect the refit
        step, which will always raise the error. Default is ``np.nan``.

    return_train_score : boolean, optional (default=False)
        If ``False``, the ``cv_results_`` attribute will not include training
        scores.

        Computing training scores is used to get insights on how different
        parameter settings impact the overfitting/underfitting trade-off.
        However computing the scores on the training set can be computationally
        expensive and is not strictly required to select the parameters that
        yield the best generalization performance.

    Examples
    --------
    >>> from sklearn.datasets import load_breast_cancer
    >>> from sklearn.tree import DecisionTreeClassifier
    >>> from sklearn.neighbors import KNeighborsClassifier
    >>> from mlresearch.model_selection import ModelSearchCV
    >>> X, y, *_ = load_breast_cancer().values()
    >>> param_grids = [{'dt__max_depth': [3, 6]}, {'kn__n_neighbors': [3, 5]}]
    >>> estimators = [('dt', DecisionTreeClassifier()), ('kn', KNeighborsClassifier())]
    >>> model_search_cv = ModelSearchCV(estimators, param_grids)
    >>> model_search_cv.fit(X, y)
    ModelSearchCV(...)
    >>> sorted(model_search_cv.cv_results_.keys())
    ['mean_fit_time', 'mean_score_time', 'mean_test_score',...]

    Attributes
    ----------
    cv_results_ : dict of numpy (masked) ndarrays
        A dict with keys as column headers and values as columns, that can be
        imported into a pandas ``DataFrame``.

        For instance the below given table

        +-------------------+-----------+------------+-----------------+---+---------+
        |param_dtc_criterion|param_gamma|param_degree|split0_test_score|...|rank_t...|
        +===================+===========+============+=================+===+=========+
        |      'entropy'    |     --    |      2     |       0.80      |...|    2    |
        +-------------------+-----------+------------+-----------------+---+---------+
        |      'entropy'    |     --    |      3     |       0.70      |...|    4    |
        +-------------------+-----------+------------+-----------------+---+---------+
        |      'entropy'    |     0.1   |     --     |       0.80      |...|    3    |
        +-------------------+-----------+------------+-----------------+---+---------+
        |      'entropy'    |     0.2   |     --     |       0.93      |...|    1    |
        +-------------------+-----------+------------+-----------------+---+---------+

        will be represented by a ``cv_results_`` dict of::

            {
            'param_kernel': masked_array(data = ['poly', 'poly', 'rbf', 'rbf'],
                                         mask = [False False False False]...)
            'param_gamma': masked_array(data = [-- -- 0.1 0.2],
                                        mask = [ True  True False False]...),
            'param_degree': masked_array(data = [2.0 3.0 -- --],
                                         mask = [False False  True  True]...),
            'split0_test_score'  : [0.80, 0.70, 0.80, 0.93],
            'split1_test_score'  : [0.82, 0.50, 0.70, 0.78],
            'mean_test_score'    : [0.81, 0.60, 0.75, 0.85],
            'std_test_score'     : [0.01, 0.10, 0.05, 0.08],
            'rank_test_score'    : [2, 4, 3, 1],
            'split0_train_score' : [0.80, 0.92, 0.70, 0.93],
            'split1_train_score' : [0.82, 0.55, 0.70, 0.87],
            'mean_train_score'   : [0.81, 0.74, 0.70, 0.90],
            'std_train_score'    : [0.01, 0.19, 0.00, 0.03],
            'mean_fit_time'      : [0.73, 0.63, 0.43, 0.49],
            'std_fit_time'       : [0.01, 0.02, 0.01, 0.01],
            'mean_score_time'    : [0.01, 0.06, 0.04, 0.04],
            'std_score_time'     : [0.00, 0.00, 0.00, 0.01],
            'params'             : [{'kernel': 'poly', 'degree': 2}, ...],
            }

        NOTE

        The key ``'params'`` is used to store a list of parameter
        settings dicts for all the parameter candidates.

        The ``mean_fit_time``, ``std_fit_time``, ``mean_score_time`` and
        ``std_score_time`` are all in seconds.

        For multi-metric evaluation, the scores for all the scorers are
        available in the ``cv_results_`` dict at the keys ending with that
        scorer's name (``'_<scorer_name>'``) instead of ``'_score'`` shown
        above. ('split0_test_precision', 'mean_train_precision' etc.)

    best_estimator_ : estimator or dict
        Estimator that was chosen by the search, i.e. estimator
        which gave highest score (or smallest loss if specified)
        on the left out data. Not available if ``refit=False``.

        See ``refit`` parameter for more information on allowed values.

    best_score_ : float
        Mean cross-validated score of the best_estimator

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

    best_params_ : dict
        Parameter setting that gave the best results on the hold out data.

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

    best_index_ : int
        The index (of the ``cv_results_`` arrays) which corresponds to the best
        candidate parameter setting.

        The dict at ``search.cv_results_['params'][search.best_index_]`` gives
        the parameter setting for the best model, that gives the highest
        mean score (``search.best_score_``).

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

    scorer_ : function or a dict
        Scorer function used on the held out data to choose the best
        parameters for the model.

        For multi-metric evaluation, this attribute holds the validated
        ``scoring`` dict which maps the scorer key to the scorer callable.

    n_splits_ : int
        The number of cross-validation splits (folds/iterations).

    refit_time_ : float
        Seconds used for refitting the best model on the whole dataset.

        This is present only if ``refit`` is not False.

    Notes
    -----
    The parameters selected are those that maximize the score of the left out
    data, unless an explicit score is passed in which case it is used instead.

    If `n_jobs` was set to a value higher than one, the data is copied for each
    point in the grid (and not `n_jobs` times). This is done for efficiency
    reasons if individual jobs take very little time, but may raise errors if
    the dataset is large and not enough memory is available.  A workaround in
    this case is to set `pre_dispatch`. Then, the memory is copied only
    `pre_dispatch` many times. A reasonable value for `pre_dispatch` is `2 *
    n_jobs`.

    """

    def __init__(
        self,
        estimators,
        param_grids,
        scoring=None,
        n_jobs=None,
        refit=True,
        cv=5,
        verbose=0,
        pre_dispatch="2*n_jobs",
        error_score="raise",
        return_train_score=False,
    ):
        estimator = (
            MultiClassifier(estimators)
            if check_estimator_type(estimators) == "classifier"
            else MultiRegressor(estimators)
        )
        est_names = [est_name for est_name, _ in estimators]
        param_grid = check_param_grids(param_grids, est_names)
        super(ModelSearchCV, self).__init__(
            estimator=estimator,
            param_grid=param_grid,
            scoring=scoring,
            n_jobs=n_jobs,
            refit=refit,
            cv=cv,
            verbose=verbose,
            pre_dispatch=pre_dispatch,
            error_score=error_score,
            return_train_score=return_train_score,
        )
        self.estimators = estimators
        self.param_grids = param_grids

    def fit(self, X, y=None, groups=None, **fit_params):

        # Call superclass fit method
        super(ModelSearchCV, self).fit(X, y, groups=groups, **fit_params)

        # Recreate best estimator attribute
        if hasattr(self, "best_estimator_"):
            self.best_estimator_ = self.best_estimator_.estimator_

        return self



from os.path import join, dirname, abspath
import types


def _optional_import(module: str) -> types.ModuleType:
    """
    Import an optional dependency.

    Parameters
    ----------
    module : str
        The identifier for the backend. Either an entrypoint item registered
        with importlib.metadata, "matplotlib", or a module name.

    Returns
    -------
    types.ModuleType
        The imported backend.
    """
    # This function was adapted from the _load_backend function from the pandas.plotting
    # source code.
    import importlib

    # Attempt an import of an optional dependency here and raise an ImportError if
    # needed.
    try:
        module_ = importlib.import_module(module)
    except ImportError:
        mod = module.split(".")[0]
        raise ImportError(f"{mod} is required to use this functionality.") from None

    return module_


def generate_paths(filepath):
    """
    Generate data, results and analysis paths.
    """
    prefix_path = join(dirname(abspath(filepath)), "..")
    paths = [join(prefix_path, name) for name in ("data", "results", "analysis")]
    return paths



import pytest
from .._utils import _optional_import


def test_optional_import():
    """Test if function ``_optional_import`` is working as expected."""
    np = _optional_import("numpy")
    assert np.__name__ == "numpy"

    exp_err_msg = "this_package_does_not_exist is required to use this functionality."
    with pytest.raises(ImportError) as err_msg:
        _optional_import("this_package_does_not_exist")

    assert err_msg.value.args[0] == exp_err_msg



"""
Implementation of the 3 synthetic data quality metrics from the paper 'How Faithful is
your Synthetic Data? Sample-level Metrics for Evaluating and Auditing Generative Models'
from Alaa et al (2022).
"""

import numpy as np
from sklearn.metrics._scorer import _BaseScorer
from sklearn.neighbors import NearestNeighbors


class _BaseSynthQualityScorer(_BaseScorer):
    def __repr__(self):
        kwargs_string = "".join([f", {k}={v}" for k, v in self.__dict__.items()])
        return f"make_scorer({self.__class__.__name__}{kwargs_string})"

    def set_score_request(self):
        """
        Placeholder to overwrite sklearn's ``_BaseScorer.set_score_request`` function.
        It is not used and was raising a docstring error with scikit-learn v1.3.0.

        Note
        ----
        This placeholder will be removed soon
        """
        pass


class AlphaPrecision(_BaseSynthQualityScorer):
    """
    Measures synthetic data fidelity. It estimates the probability that a synthetic
    sample resides in the $\\alpha$-support of the real distribution.

    This is an implementation of the metric proposed in [1]_.

    .. warning::
        This metric is not listed in the ``get_scorer_names`` function since it is
        following an unconventional structure.

    Parameters
    ----------
    scorer_real : function
        Method used to map a dataset into a score, or a 1-dimensional projection of
        itself. The mapping should be modelled over the original (real) dataset.

    alpha : float, default=0.05
        Percentile used to determine the radius of the euclidean ball.

    Attributes
    ----------
    center_ : float
        Value of the center of the euclidean ball.

    References
    ----------

    .. [1] Alaa, A., Van Breugel, B., Saveliev, E. S., & van der Schaar, M. (2022, June).
        How faithful is your synthetic data? sample-level metrics for evaluating and
        auditing generative models. In International Conference on Machine Learning
        (pp. 290-306). PMLR.

    """

    def __init__(self, scorer_real, alpha=0.05):
        self.scorer_real = scorer_real
        self.alpha = alpha

    def fit(self, X_real):
        """
        Compute statistics necessary to calculate $\\alpha$-precision.

        Parameters
        ----------
        X_real : array-like or pd.DataFrame, shape (n_samples, n_features)
            The real (original) dataset used to fit `self.scorer_real`.

        Returns
        -------
        self : object
            Returns an instance of the class.
        """
        original_scores = self.scorer_real(X_real)
        self.center_ = np.median(original_scores)
        self._dist = np.abs(original_scores - self.center_)
        return self

    def score(self, X_synth):
        """
        Returns 1 if a sample resides in the $\\alpha$-support of the original
        distribution, 0 otherwise.

        Parameters
        ----------
        X : array-like or pd.DataFrame, shape (n_samples, n_features)
            Input data over which $\\alpha$-precision will be calculated.

        Returns
        -------
        scores : np.ndarray, shape (n_samples,)
            $\\alpha$-precision scores.
        """
        radius = np.quantile(self._dist, 1 - self.alpha)
        within_ball = np.abs(self.scorer_real(X_synth) - self.center_) < radius
        return within_ball.astype(int)


class BetaRecall(_BaseSynthQualityScorer):
    """
    Checks whether the synthetic data is diverse enough to cover the variability of real
    data, i.e., a model should be able to generate a wide variety of good samples.

    This is an implementation of the metric proposed in [1]_.

    .. warning::
        This metric is not listed in the ``get_scorer_names`` function since it is
        following an unconventional structure.

    Parameters
    ----------
    scorer_synth : function
        Method used to map a dataset into a score, or a 1-dimensional projection of
        itself. The mapping should be modelled over the synthetic dataset.

    beta : float, default=0.05
        Percentile used to determine the radius of the euclidean ball.

    n_neighbors : int, default=5
        Number of neighbors to use by default for computing the radius for each sample
        in `X_real` for scoring. Ignored if `scorer_synth` is not `None`.

    metric : str or callable, default='euclidean'
        Metric to use for distance computation. Default is "euclidean", which
        results in the standard Euclidean distance. See the
        documentation of `scipy.spatial.distance
        <https://docs.scipy.org/doc/scipy/reference/spatial.distance.html>`_ and
        the metrics listed in
        :class:`~sklearn.metrics.pairwise.distance_metrics` for valid metric
        values.

        If metric is a callable function, it takes two arrays representing 1D
        vectors as inputs and must return one value indicating the distance
        between those vectors. This works for Scipy's metrics, but is less
        efficient than passing the metric name as a string.

        Ignored if `scorer_synth` is not `None`.

    n_jobs : int, default=None
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. Ignored if `scorer_synth` is not `None`.

    Attributes
    ----------
    center_ : float
        Value of the center of the euclidean ball.

    References
    ----------

    .. [1] Alaa, A., Van Breugel, B., Saveliev, E. S., & van der Schaar, M. (2022, June).
        How faithful is your synthetic data? sample-level metrics for evaluating and
        auditing generative models. In International Conference on Machine Learning
        (pp. 290-306). PMLR.

    """

    def __init__(
        self,
        scorer_synth=None,
        beta=0.05,
        n_neighbors=5,
        metric="euclidean",
        n_jobs=None,
    ):
        self.scorer_synth = scorer_synth
        self.beta = beta
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.n_jobs = n_jobs

    def _fit_with_knn(self, X_synth):
        self.center_ = X_synth.mean(axis=0)
        self._dist = np.linalg.norm(X_synth - self.center_, axis=1)

        self.radius_ = np.quantile(self._dist, 1 - self.beta)
        within_ball = np.linalg.norm(X_synth - self.center_, axis=1) < self.radius_
        self.X_synth_ = X_synth[within_ball].copy()
        return self

    def _fit_with_support_estimation(self, X_synth):
        original_scores = self.scorer_synth(X_synth).reshape(-1, 1)
        self.center_ = np.median(original_scores, axis=0)
        self._dist = np.linalg.norm(original_scores - self.center_, axis=1)

        self.radius_ = np.quantile(self._dist, 1 - self.beta)
        return self

    def fit(self, X_synth):
        """
        Compute statistics necessary to calculate $\\beta$-recall.

        Parameters
        ----------
        X_synth : array-like or pd.DataFrame, shape (n_samples, n_features)
            The synthetic dataset used to fit `self.scorer_synth`.

        Returns
        -------
        self : object
            Returns an instance of the class.
        """
        if self.scorer_synth is None:
            self._fit_with_knn(X_synth)
        else:
            self._fit_with_support_estimation(X_synth)
        return self

    def _score_with_knn(self, X_real):
        nn_real_ = NearestNeighbors(
            n_neighbors=self.n_neighbors, metric=self.metric, n_jobs=self.n_jobs
        ).fit(X_real)

        radius = nn_real_.kneighbors(X_real)[1][:, -1]

        nn_synth_ = NearestNeighbors(
            n_neighbors=1, metric=self.metric, n_jobs=self.n_jobs
        ).fit(self.X_synth_)
        dists = nn_synth_.kneighbors(X_real)[1][:, -1]

        return dists < radius

    def _score_with_support_estimation(self, X_real):
        scores = self.scorer_synth(X_real).reshape(-1, 1)
        within_ball = np.linalg.norm(scores - self.center_, axis=1) < self.radius_
        return within_ball

    def score(self, X_real):
        """
        Returns 1 if a sample resides in the $\\beta$-support of the synthetic
        distribution, 0 otherwise.

        Parameters
        ----------
        X_real : array-like or pd.DataFrame, shape (n_samples, n_features)
            Input data over which $\\beta$-recall will be calculated.

        Returns
        -------
        scores : np.ndarray, shape (n_samples,)
            $\\beta$-recall scores.
        """
        if self.scorer_synth is None:
            scores = self._score_with_knn(X_real)
        else:
            scores = self._score_with_support_estimation(X_real)
        return scores.astype(int)


class Authenticity(_BaseSynthQualityScorer):
    """
    Quantifies the rate by which a model generates new samples. In other words, this
    scorer assesses whether a sample is non-memorized.

    This is an implementation of the metric proposed in [1]_.

    .. warning::
        This metric is not listed in the ``get_scorer_names`` function since it is
        following an unconventional structure.

    Parameters
    ----------
    metric : str or callable, default='euclidean'
        Metric to use for distance computation. Default is "euclidean", which
        results in the standard Euclidean distance. See the
        documentation of `scipy.spatial.distance
        <https://docs.scipy.org/doc/scipy/reference/spatial.distance.html>`_ and
        the metrics listed in
        :class:`~sklearn.metrics.pairwise.distance_metrics` for valid metric
        values.

        If metric is a callable function, it takes two arrays representing 1D
        vectors as inputs and must return one value indicating the distance
        between those vectors. This works for Scipy's metrics, but is less
        efficient than passing the metric name as a string.

    n_jobs : int, default=None
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.

    Attributes
    ----------
    nn_ : estimator object
        Validated k-nearest neighbours algorithm. Used to find the nearest neighbors of
        the synthetic and the original data using the original data as a reference.

    distances_real_ : np.ndarray, shape (n_samples,)
        Distance to the nearest neighbor for each sample in `X_real`.

    References
    ----------

    .. [1] Alaa, A., Van Breugel, B., Saveliev, E. S., & van der Schaar, M. (2022, June).
        How faithful is your synthetic data? sample-level metrics for evaluating and
        auditing generative models. In International Conference on Machine Learning
        (pp. 290-306). PMLR.

    """

    def __init__(self, metric="euclidean", n_jobs=None):
        self.metric = metric
        self.n_jobs = n_jobs

    def fit(self, X_real):
        """
        Compute statistics necessary to calculate Authenticity.

        Parameters
        ----------
        X_real : array-like or pd.DataFrame, shape (n_samples, n_features)
            The real (original) dataset used to fit `self.scorer_real`.

        Returns
        -------
        self : object
            Returns an instance of the class.
        """
        self.nn_ = NearestNeighbors(
            n_neighbors=2, metric=self.metric, n_jobs=self.n_jobs
        ).fit(X_real)
        distances, neighbors = self.nn_.kneighbors(X_real)
        self.distances_real_ = distances[:, 1]
        return self

    def score(self, X):
        """
        Returns 1 if an observation is deemed authentic, 0 otherwise.

        Parameters
        ----------
        X : array-like or pd.DataFrame, shape (n_samples, n_features)
            Input data over which Authenticity will be calculated.

        Returns
        -------
        scores : np.ndarray, shape (n_samples,)
            Authenticity scores.
        """
        distances, neighbors = self.nn_.kneighbors(X)
        distances = distances[:, 0]
        neighbors = neighbors[:, 0]
        a_j = 1 - (distances < self.distances_real_[neighbors]).astype(int)
        return a_j
